/*
	sprec.h

	Speech analysis and recognition;
	Class PatternDB for managing a Pattern database

	(c) 1996 by Martin Pitt

	This software is distributed under the terms of the
	"GNU General Library Public License" (see file COPYING.LIB).
	
	If you perform modifications to	the sources I would appreciate
	if you sent them to me. I would also like to hear your opinion
	or improvement proposals.

	Martin Pitt
	eMail: martin.pitt@gmx.net
	home:  http://members.xoom.com/martinpitt
	*/

#ifndef sprec_h
#define sprec_h

#include <vector>
#include <string>
#include <stdio.h> // FILE*

enum PDBIOResult {
    PDB_OK,		// no error (load+save)
    PDB_CREATE,		// file cannot be created (save)
    PDB_DISKFULL,	// write error (save)
    PDB_NOTFOUND,	// file not found (load)
    PDB_EOF,		// unexpected end of file (load)
    PDB_NOPDB		// format error in PDB file (load)
};

const unsigned PATREFUSED = (unsigned) -1;
	
class Pattern {
 public:
    typedef unsigned char Element;
    typedef Element* Vector;
    
    Pattern (unsigned _vectorSize, const string& _name = "");
    ~Pattern ();

    PDBIOResult Load (FILE*);
    PDBIOResult Save (FILE*) const;

    void SetName (const string& n)	{ name = n; }
    string Name () const		{ return name; }
    unsigned VectorSize () const	{ return vectorSize; }
    unsigned DataLength () const	{ return dataLen; }
    unsigned nVectors () const		{ return dataLen/vectorSize; }
    
    long Reserved1 () const		{ return reserved1; }
    void SetReserved1 (long r)		{ reserved1 = r; }
    long Reserved2 () const		{ return reserved2; }
    void SetReserved2 (long r)		{ reserved2 = r; }

    Element operator () (unsigned vect, unsigned element) const;
    Vector AccessVector (unsigned vect);
    void NewData (unsigned _dataLen);

 private:
    string name;
    unsigned vectorSize;
    unsigned long dataLen;
    long reserved1, reserved2; // can be used freely
    Vector data;
};

class PatternDB
{    
 public:
    PatternDB (const string&);			 // create pdb
    PatternDB (const string& fName, PDBIOResult&); // load pdb
    virtual ~PatternDB ();
    
    PDBIOResult Save (const string& fName);
    
    unsigned Add (Pattern*);
    void Del (unsigned);
    void Del (Pattern*);
    void Del (const string&);
    
	//* calc distance between given Pattern and the one with index idx
    virtual unsigned Distance (Pattern&, unsigned idx) const;
    
	//* search Pattern (index) having smallest distance to p;
	//* returns distances to all Patterns via diffs (min. nPatterns()
	//* elements) if diffs != NULL
    virtual unsigned Recog (Pattern& p, unsigned *diffs = NULL) const;
    
    unsigned nPatterns () const { return pats.size (); }
    void SetName (const string& n) { name = n; }
    string Name () const { return name; }
    void SetRecogThreshold (unsigned aThreshold) {rcgThreshold=aThreshold;}
    unsigned RecogThreshold () const { return rcgThreshold; }
    void* Reserved () { return reserved; }

    Pattern& operator [] (unsigned idx) const { return *(pats[idx]); }
    
 protected:
	// called after Add(); override if you use the reserved fields;
	// gets pattern and assigned index; this one just sets the
	// reserved fields to 0.
    virtual void SetPatReserved (Pattern*, unsigned);

 protected:
    string name;
    void *reserved; // ptr to reserved 16 bytes in file header for
		    // free use
    vector <Pattern*> pats;
    unsigned rcgThreshold;
};

class SampleAnalyzer 
{
 public:
    typedef unsigned char* Sample;
    
 public:
    SampleAnalyzer (unsigned _vectorSize,
		    unsigned _windowSize = 256,
		    unsigned _sampleRate = 11025);
    
    unsigned VectorSize () const { return vectorSize; }
    unsigned WindowSize () const { return windowSize; }
    unsigned SampleRate () const { return sampleRate; }

    static unsigned Energy (Sample, unsigned size);
    Pattern* Analyze (Sample, unsigned size);
    virtual void AnalyzeWindow (Sample, Pattern::Vector) = 0;

 private:
    const unsigned vectorSize, windowSize, sampleRate;
};

class ZeroCrossAnalyzer : public SampleAnalyzer 
{
 public:
    ZeroCrossAnalyzer (unsigned _windowSize = 256,
		       unsigned _sampleRate = 11025);
    
    static unsigned ZeroCrossings (Sample, unsigned size);
    virtual void AnalyzeWindow (Sample, Pattern::Vector);

 private:
    void PartialBlockZeroHist (Sample, Pattern::Vector, char StartFrq,
			       char EndFrq);
};

class CepstrumAnalyzer : public SampleAnalyzer 
{
 public:
    CepstrumAnalyzer (unsigned _windowSize = 256,
		      unsigned _sampleRate = 11025);
    
    virtual void AnalyzeWindow (Sample, Pattern::Vector);

 private:
    char cepLevel;
};

/*****
  
  inline methods
  
  *****/

inline
Pattern::Element
Pattern::operator () (unsigned vect, unsigned elem) const
{
    return data[vect * vectorSize + elem];
}

inline
Pattern::Vector
Pattern::AccessVector (unsigned vect)
{
    return data+vect*vectorSize;
}

#endif // !defined (sprec_h)
