/*
	sound.cc  -  implementation of sound.h

	(c) 1998 Martin Pitt as Public Domain
	use, modify and spread freely!

	Version: 1.0
*/

#include "sound.h"
#include <unistd.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <sys/soundcard.h>
#include <stdlib.h>	// abs()

SoundError snd_errno = seOK;
int fAudio = 0;

inline
int
hdr2format (WAVHeader *h)
{
    if (h->nBlockAlign / h->nChannels == 1) // 8 bit
	return AFMT_U8;
    else
	return AFMT_S16_LE;
}

int
snd_initDevice (int access, unsigned freq, bool stereo, int format, WORD blockSize)
{
    int f;
    const int MAXFREQDIF = 100;

    snd_errno = seINITSOUNDCARD;

	//* open sound device ********************
    if ((fAudio = open ("/dev/dsp", access, 0)) == -1)
	return -1;

	//* set buffer size **********************
    if (blockSize)
     {
	 int b = (int) blockSize | 0x40000;
	 if (ioctl (fAudio, SNDCTL_DSP_SETFRAGMENT, &b) == -1)
	     return -1;
     }

	//* set/check sample format (8/16/mu-law) ****
    f = format;
    if (ioctl (fAudio, SNDCTL_DSP_SETFMT, &f) == -1)
	return -1;
    if (f != format)
     {
	 snd_errno = seFORMATNOTPLAYABLE;
	 return -1;
     }

	//* set/check mono/stereo ********************
    f = stereo;
    if (ioctl (fAudio, SNDCTL_DSP_STEREO, &f) == -1)
	return -1;    
    if (f != stereo)
     {
	 snd_errno = seFORMATNOTPLAYABLE;
	 return -1;
     }

	//* set/check sampling frequency *************
    f = (int) freq;
    if (ioctl (fAudio, SNDCTL_DSP_SPEED, &f) == -1)
	return -1;
    if (abs ((unsigned) f - freq) > MAXFREQDIF)
     {
	 snd_errno = seFORMATNOTPLAYABLE;
	 return -1;
     }

    snd_errno = seOK;
    return fAudio;
}

int
snd_playWAV (const char* fname)
{
    int fWAV;
    WAVHeader h;
    char *data;
    ULONG bytesToRead;
    int bs;

    if ((fWAV = open (fname, O_RDONLY, 0)) == -1)
     {
	 snd_errno = seFILENOTFOUND;
	 return -1;
     }

	//* read header ******************

    if (read (fWAV, &h, sizeof (h)) != sizeof (h))
     {
	 snd_errno = seFILEFORMAT;
	 return -1;
     }
    
	//* check header ********************

    if (memcmp (h.rID, "RIFF", 4) || memcmp (h.wID, "WAVE", 4) ||
	memcmp (h.fID, "fmt ", 4) || memcmp (h.dID, "data", 4))
     {
	 snd_errno = seFILEFORMAT;
	 return -1;
     }

    if (h.wFormatTag != 1)
     {
	 snd_errno = seFORMATUNKNOWN;
	 return -1;
     }

	//* prepare playing ****************

    if (snd_initDevice (O_WRONLY, h.nSamplesPerSec, h.nChannels == 2,
			hdr2format (&h), 12) == -1)
	return -1;

    data = new char[4096];
    bytesToRead = h.dLen;
    bs = 4096;

	//* play loop **********************
    while (bytesToRead)
     {
	 if (bytesToRead < 4096)
	  {
	      bs = bytesToRead;
	      bytesToRead = 0;
	  }
	 else
	      bytesToRead -= 4096;
	 
	 if (read (fWAV, data, bs) != bs)
	  {
	      snd_errno = seUNEXPECTEDEOF;
	      close (fAudio);
	      fAudio = 0;
	      close (fWAV);
	      delete data;
	      return -1;
	  }

	 write (fAudio, data, bs);
     }
    
	//* shutdown ***********************
    close (fAudio);
    fAudio = 0;
    
    close (fWAV);
    delete data;
    
    return 0;
}

const char*
snd_strerror ()
{
    switch (snd_errno)
     {
	 case seOK : return "no error";
	 case seFILENOTFOUND: return "file not found";
	 case seFILEFORMAT: return "file has invalid format";
	 case seINITSOUNDCARD: return "soundcard could not be initialized";
	 case seFORMATUNKNOWN: return "function cannot handle this format";
	 case seFORMATNOTPLAYABLE: return "soundcard is not capable of handling this format";
	 case seUNEXPECTEDEOF: return "file shorter than stated in its header";
     }

    return "";
}
