/*
	sprec.cc

	(c) 1996 by Martin Pitt
*/

#include "sprec.h"

#include <sndxform.h>
#include <string.h> // memset(), memcpy()
#include <algo.h> // find()

const char PDBMAGIC[] = "Speech Pattern Database\x1B";

typedef struct {
    char magic[24];	// should contain PDBMAGIC
    char name[30];      // name of data base
    unsigned num,	// number of contained Patterns
	rcgThreshold;	// refuse threshold
    char reserved[16];	// reserved data for user's purposes
} PDBFileHeader;

/*****
  
  Pattern
  
  *****/

Pattern::Pattern (unsigned _vectorSize, const string& _name) 
	: name (_name), vectorSize (_vectorSize), dataLen (0),
	  reserved1 (0), reserved2 (0), data (NULL)
{}

Pattern::~Pattern ()
{
    if (data)
	delete data;
}

PDBIOResult
Pattern::Load (FILE* f) 
{
	//* read name **
    unsigned l;
    char* n;
    if (fread (&l, sizeof (l), 1, f) != 1)
	return PDB_EOF;
    n = new char[l+1];
    n[l] = 0;
    if (fread (n, l, 1, f) != 1)
	return PDB_EOF;
    name = n;
    delete n;

	//* read other fields
    fread (&vectorSize, sizeof (vectorSize), 1, f);
    fread (&dataLen, sizeof (dataLen), 1, f);
    fread (&reserved1, sizeof (reserved1), 1, f);
    fread (&reserved2, sizeof (reserved2), 1, f);
    NewData (dataLen);
    if (fread (data, dataLen, 1, f) != 1)
	return PDB_EOF;

    return PDB_OK;
}

PDBIOResult
Pattern::Save (FILE* f) const
{
    unsigned l = name.length ();
    if ( (fwrite (&l, sizeof (l), 1, f) != 1) ||
	 (fwrite (name.c_str(), l, 1, f) != 1) ||
	 (fwrite (&vectorSize, sizeof (vectorSize), 1, f) != 1) ||
	 (fwrite (&dataLen, sizeof (dataLen), 1, f) != 1) ||
	 (fwrite (&reserved1, sizeof (reserved1), 1, f) != 1) ||
	 (fwrite (&reserved2, sizeof (reserved2), 1, f) != 1) ||
	 (fwrite (data, dataLen, 1, f) != 1) )
	return PDB_DISKFULL;
    
    return PDB_OK;
}

void
Pattern::NewData (unsigned _dataLen)
{
    if (data)
	delete data;
    if ((dataLen = _dataLen) != 0)
	data = new Element[dataLen];
    else
	data = NULL;
}

/*****

  PatternDB

  *****/

PatternDB::PatternDB (const string& _name) :
	name (_name), pats (), rcgThreshold (120)
{
    reserved = new char[16];
    memset (reserved, 0, 16);
}

PatternDB::PatternDB (const string& fName, PDBIOResult& err) 
	: name(), pats ()
{
    FILE *f;
    PDBFileHeader hdr;

    reserved = (void*) new char[16];
    
    if ((f = fopen (fName.c_str(), "rb")) == NULL)
     {
	 err = PDB_NOTFOUND;
	 return;
     }
    
//* read header, check magic **********
    if (!fread (&hdr, sizeof (PDBFileHeader), 1, f) ||
	strncmp (PDBMAGIC, hdr.magic, sizeof (PDBMAGIC)-1))
     {
	 fclose (f);
	 err = PDB_NOPDB;
	 return;
     }

    name = hdr.name;
    rcgThreshold = hdr.rcgThreshold;
    memcpy (reserved, hdr.reserved, 16);
    
// read Patterns **************************
    for (unsigned i = 0; i < hdr.num; ++i)
     {
	 Pattern* p = new Pattern (0);
	 if ((err = p->Load (f)) != PDB_OK)
	  {
	      delete p;
	      fclose (f);
	      return;
	  }
	  else
	     pats.push_back (p);
     }
    
    fclose (f);
}

PatternDB::~PatternDB ()
{
  vector<Pattern*>::iterator i;

  for (i = pats.begin(); i != pats.end(); ++i)
      delete *i;
  delete reserved;
}

PDBIOResult
PatternDB::Save (const string& fName)
{
  FILE *f;
  PDBFileHeader hdr;

//* create file, write header ****
  if ((f = fopen (fName.c_str(), "wb")) == NULL)
    return PDB_CREATE;

  memset (&hdr, 0, sizeof (hdr));
  strcpy (hdr.magic, PDBMAGIC);
  strcpy (hdr.name, name.c_str());
  memcpy (hdr.reserved, reserved, 16);
  hdr.num = pats.size();
  hdr.rcgThreshold = rcgThreshold;

  fwrite (&hdr, sizeof (hdr), 1, f);

//* write Patterns *********************
  for (vector<Pattern*>::iterator i = pats.begin(); i != pats.end(); ++i)
      if ((*i)->Save (f) != PDB_OK)
       {
	   fclose (f);
	   return PDB_DISKFULL;
       }

  fclose (f);
  return PDB_OK;
}

unsigned
PatternDB::Add (Pattern* p)
{

//* search position to insert at (alphabetically sorted) *
    vector<Pattern*>::iterator i;
    unsigned pos = 0;
    for (i = pats.begin(); i != pats.end(); ++i, ++pos)
	if ((*i)->Name().compare (p->Name()) > 0)
	 {
	     pats.insert (i, p);
	     SetPatReserved (p, pos);
	     return pos;
	 }
    
    pats.push_back (p);
    SetPatReserved (p, ++pos);
    return pos;
}

void
PatternDB::Del (Pattern* p)
{
    pats.erase (find (pats.begin(), pats.end(), p));
}

void
PatternDB::Del (const string& id)
{
    vector<Pattern*>::iterator i;
    for (i = pats.begin(); i != pats.end(); ++i)
	if ((*i)->Name() == id)
	 {
	     delete *i;
	     pats.erase (i);
	     return;
	 }
}

unsigned
minOf3 (unsigned x1, unsigned x2, unsigned x3)
{
    return (x1<x2) ? ((x1<x3) ? x1 : x3) : ((x2<x3) ? x2 : x3);
}

// distance between 2 vectors: d = |v1-v2|^2
unsigned
vecDist (Pattern::Vector v1, Pattern::Vector v2, unsigned size)
{
    unsigned diff = 0;
    for (unsigned i = 0; i < size; ++i)
	diff += (signed char) (v1[i] - v2[i]) * (signed char) (v1[i] - v2[i]);
    return diff;
}

unsigned
PatternDB::Distance (Pattern &p, unsigned idx) const
{
    unsigned *currpath = new unsigned [p.nVectors()];
    unsigned *nextrow = new unsigned [p.nVectors()];
    
    *currpath = 0;
    for (unsigned row=0; row < pats[idx]->nVectors(); ++row)
     {
	 for (unsigned col = 0; col < p.nVectors(); ++col)
	  {
		  // find best continuation from recent path
	      if (col == 0) // leftmost element can come only from below
		  *nextrow = *currpath;
	       else if (row == 0)  // bottom element can come only from left
		   nextrow[col] = nextrow[col-1];
	       else
		   nextrow[col] = minOf3 (currpath[col], currpath[col-1], nextrow[col-1]);
	 
		  // add local distance
	      nextrow[col] += vecDist (p.AccessVector (col),
				       pats[idx]->AccessVector (row),
				       p.VectorSize());
	  }
	 
	     // advance to next row
	 unsigned* swap = currpath;
	 currpath = nextrow;
	 nextrow = swap;
     }

    unsigned d = currpath[p.nVectors()-1];
    delete currpath;
    delete nextrow;
    return d;
}

unsigned
PatternDB::Recog (Pattern& p, unsigned *diffs) const
{
    unsigned diff, min = PATREFUSED, recg = PATREFUSED;
    
    for (unsigned i = 0; i < pats.size(); ++i)
     {
	     //* calc and normalize distance ***
	 unsigned dist = Distance (p, i);
	 diff = (unsigned) ((unsigned long) dist*200/pats[i]->DataLength());
	 
	 if (diffs)
	     diffs[i] = diff;
	 
	 if (diff < min)
	  {
	      min = diff;
	      recg = i;
	  }
     }
    
    return (min < rcgThreshold) ? recg : PATREFUSED;
}

void
PatternDB::SetPatReserved (Pattern *p, unsigned)
{
    p->SetReserved1 (0);
    p->SetReserved2 (0);
}

/*****

  SampleAnalyzer

  *****/

SampleAnalyzer::SampleAnalyzer (unsigned _vectorSize, unsigned
				_windowSize, unsigned _sampleRate) 
	: vectorSize (_vectorSize), windowSize (_windowSize),
	  sampleRate (_sampleRate)
{}

unsigned
SampleAnalyzer::Energy (Sample sample, unsigned size)
{
    unsigned erg = 0, i;
    signed char currSample;
    
    for (i = 0; i < size; i++)
     {
	 currSample = (signed char) sample[i] - 128;
	 erg += currSample * currSample;
     }
    return erg;
}


Pattern*
SampleAnalyzer::Analyze (Sample sample, unsigned size)
{
    unsigned i, v;
    Pattern* p = new Pattern (VectorSize());
    p->NewData ((size/WindowSize()+1)*VectorSize());
    
    for (i = 0, v = 0; i < size; i += WindowSize(), ++v)
	AnalyzeWindow (sample+i, p->AccessVector (v));

    return p;
}

/*****

  ZeroCrossAnalyzer

  *****/

ZeroCrossAnalyzer::ZeroCrossAnalyzer (unsigned _windowSize,
				      unsigned _sampleRate) 
	: SampleAnalyzer (8, _windowSize, _sampleRate)
{}

unsigned
ZeroCrossAnalyzer::ZeroCrossings (Sample sample, unsigned size)
{
    unsigned i, crs = 0;
    for (i = size-1; i; --i)
	crs += (sample[i] & 0x80) ^ (sample[i-1] & 0x80);
    return crs;
}

void
ZeroCrossAnalyzer::AnalyzeWindow (Sample sample, Pattern::Vector destVect)
{
    Sample filtered = new unsigned char[WindowSize ()];
    static const float divs[7] = {2, 3.5, 3.6, 3.8, 4, 12, 25};
    
    memset (destVect, 0, VectorSize());
    
//* filter and analyze *********************
	// Tiefpass f=11025 Hz, fMid=475 Hz, fWidth=950 Hz
    sampleMedPassX (sample, filtered, WindowSize(), 11, 147, -58);
    PartialBlockZeroHist (filtered, destVect, 0, 3);
    
    sampleHiPass (sample, filtered, WindowSize());
    PartialBlockZeroHist (filtered, destVect, 4, 7);
    
    delete filtered;
    
	// normalize all values to range 0...10 (required because higher
	// frequencies cause higher crossing rates)
    for (unsigned i = 7; i; i--)
	destVect[i] = (Pattern::Element) (destVect[i]/divs[i-1]+0.5);
}

void
ZeroCrossAnalyzer::PartialBlockZeroHist (Sample src, Sample dest,
					 char StartFrq, char EndFrq)
{
	//* frequency bank boundaries *
    static const unsigned fBounds[8]={200, 400, 700, 1000, 1400, 1800,
				      2600, 5000};
    unsigned lastCrs = 0, i;
    char frq;
    
    for (i = 1; i < WindowSize(); ++i)
	if ((src[i]^src[i-1]) & 0x80)
	 {
	     frq = StartFrq;
	     while ((i-lastCrs < SampleRate()/fBounds[frq]) && (frq <= EndFrq))
		 ++frq;
	     if (frq > EndFrq)
		 continue;
	     
	     ++dest[frq];
	     lastCrs = i;
	 }
}


/*****

  CepstrumAnalyzer

  *****/

CepstrumAnalyzer::CepstrumAnalyzer (unsigned _windowSize,
				    unsigned _sampleRate) 
	: SampleAnalyzer (16, _windowSize, _sampleRate)
{
    switch (WindowSize ())
     {
	 case 128: cepLevel = 7; break;
	 case 256: cepLevel = 8; break;
	 case 512: cepLevel = 9; break;
	 case 1024: cepLevel = 10; break;
	 default: cepLevel = 0;
     }
}

void
CepstrumAnalyzer::AnalyzeWindow (Sample sample, Pattern::Vector destVect)
{
    cepstrum (sample, cepLevel, destVect, 4, VectorSize(), 0.1);
}
